from dataclasses import dataclass, field
from typing import Any

from ruamel import yaml

from ..context import Context
from ..extractors import extract_extension_fields
from ..fields import fixed_field, fixed_fields
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .oauth_flow import OAuthFlow
from .oauth_flow import build as build_oauth_flow


__all__ = ["OAuthFlows", "build"]


@dataclass(frozen=True, slots=True)
class OAuthFlows:
    """
    OAuth Flows Object representation for OpenAPI 3.1.

    A container for the list of possible OAuth 2.0 authorization flows.
    Used within Security Scheme Objects when type is set to "oauth2".

    Attributes:
        root_node: The top-level node representing the entire OAuth Flows object in the original source file
        implicit: Configuration for the OAuth Implicit flow
        password: Configuration for the OAuth Resource Owner Password flow
        client_credentials: Configuration for the OAuth Client Credentials flow (application flow)
        authorization_code: Configuration for the OAuth Authorization Code flow (three-legged OAuth)
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    implicit: FieldSource[OAuthFlow] | None = fixed_field()
    password: FieldSource[OAuthFlow] | None = fixed_field()
    client_credentials: FieldSource[OAuthFlow] | None = fixed_field(
        metadata={"yaml_name": "clientCredentials"}
    )
    authorization_code: FieldSource[OAuthFlow] | None = fixed_field(
        metadata={"yaml_name": "authorizationCode"}
    )
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> OAuthFlows | ValueSource[YAMLInvalidValue]:
    """
    Build an OAuthFlows object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An OAuthFlows object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("implicit:\\n  authorizationUrl: https://example.com/auth\\n  scopes: {}")
        flows = build(root)
        assert flows.implicit.value.authorization_url.value == 'https://example.com/auth'
    """
    context = context or Context()

    if not isinstance(root, yaml.MappingNode):
        # Preserve invalid root data instead of returning None
        value = context.yaml_constructor.construct_object(root, deep=True)
        return ValueSource(value=value, value_node=root)

    # Get fixed specification fields for this dataclass type
    _fixed_fields = fixed_fields(OAuthFlows)

    # Build YAML name to Python field name mapping
    yaml_to_field = {
        field.metadata.get("yaml_name", fname): fname for fname, field in _fixed_fields.items()
    }

    # Extract field values in a single pass
    field_values: dict[str, Any] = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        # Map YAML key to Python field name
        field_name = yaml_to_field.get(key)
        if field_name:
            # Build OAuthFlow for this field - child builder handles invalid nodes
            # FieldSource will auto-unwrap ValueSource if child returns it for invalid data
            oauth_flow = build_oauth_flow(value_node, context=context)
            field_values[field_name] = FieldSource(
                value=oauth_flow, key_node=key_node, value_node=value_node
            )

    return OAuthFlows(
        root_node=root,
        implicit=field_values.get("implicit"),
        password=field_values.get("password"),
        client_credentials=field_values.get("client_credentials"),
        authorization_code=field_values.get("authorization_code"),
        extensions=extract_extension_fields(root, context),
    )



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .media_type import MediaType
from .reference import Reference
from .reference import build as build_reference


__all__ = ["RequestBody", "build", "build_request_body_or_reference"]


@dataclass(frozen=True, slots=True)
class RequestBody:
    """
    Request Body Object representation for OpenAPI 3.1.

    Describes a single request body.

    Attributes:
        root_node: The top-level node representing the entire Request Body object in the original source file
        description: A brief description of the request body. CommonMark syntax MAY be used for rich text representation.
        content: The content of the request body. The key is a media type or media type range and the value describes it.
                For requests that match multiple keys, only the most specific key is applicable.
        required: Determines if the request body is required in the request. Defaults to false.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    description: FieldSource[str] | None = fixed_field()
    content: FieldSource[dict[KeySource[str], "MediaType"]] | None = fixed_field()
    required: FieldSource[bool] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> RequestBody | ValueSource[YAMLInvalidValue]:
    """
    Build a RequestBody object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A RequestBody object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        description: user to add to the system
        required: true
        content:
          application/json:
            schema:
              type: object
        ''')
        request_body = build(root)
        assert request_body.description.value == 'user to add to the system'
    """
    context = context or Context()

    # Use build_model for initial construction
    request_body = build_model(root, RequestBody, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(request_body, RequestBody):
        return request_body

    return request_body


def build_request_body_or_reference(
    node: yaml.Node, context: Context
) -> RequestBody | Reference | ValueSource[YAMLInvalidValue]:
    """
    Build either a RequestBody or Reference from a YAML node.

    This helper handles the polymorphic nature of OpenAPI where many fields
    can contain either a RequestBody object or a Reference object ($ref).

    Args:
        node: The YAML node to parse
        context: Parsing context

    Returns:
        A RequestBody, Reference, or ValueSource if the node is invalid
    """
    # Check if it's a reference (has $ref key)
    if isinstance(node, yaml.MappingNode):
        for key_node, _ in node.value:
            key = context.yaml_constructor.construct_yaml_str(key_node)
            if key == "$ref":
                return build_reference(node, context)

    # Otherwise, try to build as RequestBody
    return build(node, context)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .example import Example
from .reference import Reference
from .reference import build as build_reference
from .schema import BooleanJSONSchema, Schema


if TYPE_CHECKING:
    from .media_type import MediaType


__all__ = ["Header", "build", "build_header_or_reference"]


@dataclass(frozen=True, slots=True)
class Header:
    """
    Header Object representation for OpenAPI 3.1.

    The Header Object follows the structure of the Parameter Object, including determining its
    serialization strategy based on whether schema or content is present, with the following changes:
    - name MUST NOT be specified, it is given in the corresponding headers map.
    - in MUST NOT be specified, it is implicitly in header.
    - All traits that are affected by the location MUST be applicable to a location of header
      (for example, style). This means that allowEmptyValue and allowReserved MUST NOT be used,
      and style, if used, MUST be limited to "simple".

    Attributes:
        root_node: The top-level node representing the entire Header object in the original source file
        description: A brief description of the header. CommonMark syntax MAY be used for rich text representation.
        required: Determines whether this header is mandatory. Default value is false.
        deprecated: Specifies that a header is deprecated and SHOULD be transitioned out of usage. Default value is false.
        style: Describes how the header value will be serialized depending on the type of the header value.
               If used, MUST be limited to "simple".
        explode: When this is true, header values of type array or object generate separate parameters for each value of the array or key-value pair of the map. Default value is false.
        schema: The schema defining the type used for the header.
        example: Example of the header's potential value. The example SHOULD match the specified schema and encoding properties if present.
        examples: Examples of the header's potential value. Each example SHOULD contain a value in the correct format as specified in the header encoding.
        content: A map containing the representations for the header. The key is the media type and the value describes it.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    description: FieldSource[str] | None = fixed_field()
    required: FieldSource[bool] | None = fixed_field()
    deprecated: FieldSource[bool] | None = fixed_field()
    style: FieldSource[str] | None = fixed_field()
    explode: FieldSource[bool] | None = fixed_field()
    schema: FieldSource["Schema | BooleanJSONSchema"] | None = fixed_field()
    example: FieldSource[YAMLValue] | None = fixed_field()
    examples: FieldSource[dict[KeySource[str], "Example | Reference"]] | None = fixed_field()
    content: FieldSource[dict[KeySource[str], "MediaType"]] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Header | ValueSource[YAMLInvalidValue]:
    """
    Build a Header object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Header object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        description: The number of allowed requests in the current period
        schema:
          type: integer
        ''')
        header = build(root)
        assert header.description.value == 'The number of allowed requests in the current period'
    """
    return build_model(root, Header, context=context)


def build_header_or_reference(
    node: yaml.Node, context: Context
) -> Header | Reference | ValueSource[YAMLInvalidValue]:
    """
    Build either a Header or Reference from a YAML node.

    This helper handles the polymorphic nature of OpenAPI where many fields
    can contain either a Header object or a Reference object ($ref).

    Args:
        node: The YAML node to parse
        context: Parsing context

    Returns:
        A Header, Reference, or ValueSource if the node is invalid
    """
    # Check if it's a reference (has $ref key)
    if isinstance(node, yaml.MappingNode):
        for key_node, _ in node.value:
            key = context.yaml_constructor.construct_yaml_str(key_node)
            if key == "$ref":
                return build_reference(node, context)

    # Otherwise, try to build as Header
    return build(node, context)



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_field_source, build_model
from .components import Components
from .components import build as build_components
from .external_documentation import ExternalDocumentation
from .info import Info
from .info import build as build_info
from .paths import Paths
from .paths import build as build_paths
from .security_requirement import SecurityRequirement
from .server import Server
from .tag import Tag
from .tag import build as build_tag


__all__ = ["OpenAPI30", "build"]


@dataclass(frozen=True, slots=True)
class OpenAPI30:
    """
    OpenAPI Object representation for OpenAPI 3.0.

    This is the root document object of the OpenAPI document.

    Attributes:
        root_node: The top-level node representing the entire OpenAPI object in the original source file
        openapi: REQUIRED. The version number of the OpenAPI Specification that the document uses.
                 Must match the specification version (e.g., "3.0.0", "3.0.1", "3.0.2", "3.0.3", "3.0.4").
        info: REQUIRED. Provides metadata about the API (title, version, description, etc.)
        paths: REQUIRED. The available paths and operations for the API
        servers: An array of Server Objects providing connectivity information to target servers
        components: An element to hold reusable objects for different aspects of the OAS
        security: A declaration of which security mechanisms can be used across the API
        tags: A list of tags used by the specification with additional metadata
        external_docs: Additional external documentation
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    openapi: FieldSource[str] | None = fixed_field()
    info: FieldSource[Info] | None = fixed_field()
    servers: FieldSource[list["Server"]] | None = fixed_field()
    paths: FieldSource[Paths] | None = fixed_field()
    components: FieldSource[Components] | None = fixed_field()
    security: FieldSource[list["SecurityRequirement"]] | None = fixed_field()
    tags: FieldSource[list[Tag]] | None = fixed_field()
    external_docs: FieldSource["ExternalDocumentation"] | None = fixed_field(
        metadata={"yaml_name": "externalDocs"}
    )
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> OpenAPI30 | ValueSource[YAMLInvalidValue]:
    """
    Build an OpenAPI object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An OpenAPI object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        openapi: 3.0.4
        info:
          title: Sample API
          version: 1.0.0
        paths:
          /users:
            get:
              responses:
                '200':
                  description: Success
        ''')
        openapi_doc = build(root)
        assert openapi_doc.openapi.value == '3.0.4'
        assert openapi_doc.info.value.title.value == 'Sample API'
    """
    context = context or Context()

    # Use build_model for initial construction
    openapi_obj = build_model(root, OpenAPI30, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(openapi_obj, OpenAPI30):
        return openapi_obj

    # Manually handle nested complex fields and list fields with custom builders
    replacements = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "info":
            # Handle info field - Info object
            replacements["info"] = FieldSource(
                value=build_info(value_node, context), key_node=key_node, value_node=value_node
            )

        elif key == "paths":
            # Handle paths field - Paths object
            replacements["paths"] = FieldSource(
                value=build_paths(value_node, context), key_node=key_node, value_node=value_node
            )

        elif key == "components":
            # Handle components field - Components object
            replacements["components"] = FieldSource(
                value=build_components(value_node, context),
                key_node=key_node,
                value_node=value_node,
            )

        elif key == "tags":
            # Handle tags field - array of Tag objects
            if isinstance(value_node, yaml.SequenceNode):
                tags_list = []
                for tag_node in value_node.value:
                    tag_obj = build_tag(tag_node, context)
                    tags_list.append(tag_obj)
                replacements["tags"] = FieldSource(
                    value=tags_list, key_node=key_node, value_node=value_node
                )
            else:
                # Not a sequence - preserve as-is for validation
                replacements["tags"] = build_field_source(key_node, value_node, context)

    # Apply all replacements at once
    if replacements:
        openapi_obj = replace(openapi_obj, **replacements)

    return openapi_obj



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .operation import Operation
from .operation import build as build_operation
from .parameter import Parameter
from .reference import Reference
from .server import Server


__all__ = ["PathItem", "build"]


@dataclass(frozen=True, slots=True)
class PathItem:
    """
    Path Item Object representation for OpenAPI 3.1.

    Describes the operations available on a single path. A Path Item may be empty,
    due to ACL constraints.

    Attributes:
        root_node: The top-level node representing the entire Path Item object in the original source file
        ref: Allows referencing an external definition of this path item
        summary: Optional string summary intended to apply to all operations in this path
        description: Optional string description (may contain CommonMark syntax)
        get: Definition of a GET operation on this path
        put: Definition of a PUT operation on this path
        post: Definition of a POST operation on this path
        delete: Definition of a DELETE operation on this path
        options: Definition of an OPTIONS operation on this path
        head: Definition of a HEAD operation on this path
        patch: Definition of a PATCH operation on this path
        trace: Definition of a TRACE operation on this path
        servers: Alternative server array to service all operations in this path
        parameters: List of parameters that are applicable for all operations in this path
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$ref"})
    summary: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    get: FieldSource[Operation] | None = fixed_field()
    put: FieldSource[Operation] | None = fixed_field()
    post: FieldSource[Operation] | None = fixed_field()
    delete: FieldSource[Operation] | None = fixed_field()
    options: FieldSource[Operation] | None = fixed_field()
    head: FieldSource[Operation] | None = fixed_field()
    patch: FieldSource[Operation] | None = fixed_field()
    trace: FieldSource[Operation] | None = fixed_field()
    servers: FieldSource[list["Server"]] | None = fixed_field()
    parameters: FieldSource[list["Parameter | Reference"]] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> PathItem | ValueSource[YAMLInvalidValue]:
    """
    Build a PathItem object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A PathItem object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        summary: User operations
        get:
          summary: List users
          responses:
            '200':
              description: successful operation
        post:
          summary: Create user
          responses:
            '201':
              description: user created
        ''')
        path_item = build(root)
        assert path_item.get is not None
        assert path_item.post is not None
    """
    context = context or Context()

    # Use build_model for initial construction
    path_item = build_model(root, PathItem, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(path_item, PathItem):
        return path_item

    # Manually handle nested complex fields
    replacements = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        # Handle HTTP method operation fields
        if key in ("get", "put", "post", "delete", "options", "head", "patch", "trace"):
            operation_obj = build_operation(value_node, context)
            replacements[key] = FieldSource(
                value=operation_obj, key_node=key_node, value_node=value_node
            )

    # Apply all replacements at once
    if replacements:
        path_item = replace(path_item, **replacements)

    return path_item



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_field_source, build_model
from .callback import Callback
from .example import Example, build_example_or_reference
from .header import Header
from .link import Link
from .parameter import Parameter, build_parameter_or_reference
from .path_item import PathItem
from .reference import Reference
from .request_body import RequestBody, build_request_body_or_reference
from .response import Response, build_response_or_reference
from .schema import BooleanJSONSchema, Schema
from .schema import build as build_schema
from .security_scheme import SecurityScheme, build_security_scheme_or_reference


__all__ = ["Components", "build"]


@dataclass(frozen=True, slots=True)
class Components:
    r"""
    Components Object representation for OpenAPI 3.1.

    Holds reusable objects for different aspects of the OAS. All objects defined within
    the components object have no effect on the API unless they are explicitly referenced
    from properties outside the components object.

    All component keys MUST match the regex pattern: ^[a-zA-Z0-9\.\-_]+$
    (Note: This is a validation concern and not enforced by this low-level model)

    Attributes:
        root_node: The top-level node representing the entire Components object in the original source file
        schemas: Reusable Schema Objects (can include $ref directly in OpenAPI 3.1)
        responses: Reusable Response Objects or Reference Objects
        parameters: Reusable Parameter Objects or Reference Objects
        examples: Reusable Example Objects or Reference Objects
        request_bodies: Reusable Request Body Objects or Reference Objects
        headers: Reusable Header Objects or Reference Objects
        security_schemes: Reusable Security Scheme Objects or Reference Objects
        links: Reusable Link Objects or Reference Objects
        callbacks: Reusable Callback Objects or Reference Objects
        path_items: Reusable Path Item Objects (new in OpenAPI 3.1)
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    schemas: FieldSource[dict[KeySource[str], Schema | BooleanJSONSchema]] | None = fixed_field()
    responses: FieldSource[dict[KeySource[str], Response | Reference]] | None = fixed_field()
    parameters: FieldSource[dict[KeySource[str], Parameter | Reference]] | None = fixed_field()
    examples: FieldSource[dict[KeySource[str], Example | Reference]] | None = fixed_field()
    request_bodies: FieldSource[dict[KeySource[str], RequestBody | Reference]] | None = fixed_field(
        metadata={"yaml_name": "requestBodies"}
    )
    headers: FieldSource[dict[KeySource[str], "Header | Reference"]] | None = fixed_field()
    security_schemes: FieldSource[dict[KeySource[str], SecurityScheme | Reference]] | None = (
        fixed_field(metadata={"yaml_name": "securitySchemes"})
    )
    links: FieldSource[dict[KeySource[str], "Link | Reference"]] | None = fixed_field()
    callbacks: FieldSource[dict[KeySource[str], "Callback | Reference"]] | None = fixed_field()
    path_items: FieldSource[dict[KeySource[str], "PathItem"]] | None = fixed_field(
        metadata={"yaml_name": "pathItems"}
    )
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Components | ValueSource[YAMLInvalidValue]:
    """
    Build a Components object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Components object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        schemas:
          User:
            type: object
            properties:
              id:
                type: integer
              name:
                type: string
        responses:
          NotFound:
            description: Entity not found
        ''')
        components = build(root)
        assert 'User' in {k.value for k in components.schemas.value.keys()}
    """
    context = context or Context()

    # Use build_model for initial construction
    components_obj = build_model(root, Components, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(components_obj, Components):
        return components_obj

    # Manually handle nested complex fields that aren't covered by build_model
    replacements = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "schemas":
            # Handle schemas field - map of Schema objects (in 3.1, schemas can have $ref directly)
            if isinstance(value_node, yaml.MappingNode):
                schemas_dict = {}
                for schema_key_node, schema_value_node in value_node.value:
                    schema_key = context.yaml_constructor.construct_yaml_str(schema_key_node)
                    schema = build_schema(schema_value_node, context)
                    schemas_dict[KeySource(value=schema_key, key_node=schema_key_node)] = schema
                replacements["schemas"] = FieldSource(
                    value=schemas_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["schemas"] = build_field_source(key_node, value_node, context)

        elif key == "responses":
            # Handle responses field - map of Response or Reference objects
            if isinstance(value_node, yaml.MappingNode):
                responses_dict = {}
                for response_key_node, response_value_node in value_node.value:
                    response_key = context.yaml_constructor.construct_yaml_str(response_key_node)
                    response_or_reference = build_response_or_reference(
                        response_value_node, context
                    )
                    responses_dict[KeySource(value=response_key, key_node=response_key_node)] = (
                        response_or_reference
                    )
                replacements["responses"] = FieldSource(
                    value=responses_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["responses"] = build_field_source(key_node, value_node, context)

        elif key == "parameters":
            # Handle parameters field - map of Parameter or Reference objects
            if isinstance(value_node, yaml.MappingNode):
                parameters_dict = {}
                for parameter_key_node, parameter_value_node in value_node.value:
                    parameter_key = context.yaml_constructor.construct_yaml_str(parameter_key_node)
                    parameter_or_reference = build_parameter_or_reference(
                        parameter_value_node, context
                    )
                    parameters_dict[KeySource(value=parameter_key, key_node=parameter_key_node)] = (
                        parameter_or_reference
                    )
                replacements["parameters"] = FieldSource(
                    value=parameters_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["parameters"] = build_field_source(key_node, value_node, context)

        elif key == "examples":
            # Handle examples field - map of Example or Reference objects
            if isinstance(value_node, yaml.MappingNode):
                examples_dict = {}
                for example_key_node, example_value_node in value_node.value:
                    example_key = context.yaml_constructor.construct_yaml_str(example_key_node)
                    example_or_reference = build_example_or_reference(example_value_node, context)
                    examples_dict[KeySource(value=example_key, key_node=example_key_node)] = (
                        example_or_reference
                    )
                replacements["examples"] = FieldSource(
                    value=examples_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["examples"] = build_field_source(key_node, value_node, context)

        elif key == "requestBodies":
            # Handle requestBodies field - map of RequestBody or Reference objects
            if isinstance(value_node, yaml.MappingNode):
                request_bodies_dict = {}
                for request_body_key_node, request_body_value_node in value_node.value:
                    request_body_key = context.yaml_constructor.construct_yaml_str(
                        request_body_key_node
                    )
                    request_body_or_reference = build_request_body_or_reference(
                        request_body_value_node, context
                    )
                    request_bodies_dict[
                        KeySource(value=request_body_key, key_node=request_body_key_node)
                    ] = request_body_or_reference
                replacements["request_bodies"] = FieldSource(
                    value=request_bodies_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["request_bodies"] = build_field_source(key_node, value_node, context)

        elif key == "securitySchemes":
            # Handle securitySchemes field - map of SecurityScheme or Reference objects
            if isinstance(value_node, yaml.MappingNode):
                security_schemes_dict = {}
                for security_scheme_key_node, security_scheme_value_node in value_node.value:
                    security_scheme_key = context.yaml_constructor.construct_yaml_str(
                        security_scheme_key_node
                    )
                    security_scheme_or_reference = build_security_scheme_or_reference(
                        security_scheme_value_node, context
                    )
                    security_schemes_dict[
                        KeySource(value=security_scheme_key, key_node=security_scheme_key_node)
                    ] = security_scheme_or_reference
                replacements["security_schemes"] = FieldSource(
                    value=security_schemes_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["security_schemes"] = build_field_source(key_node, value_node, context)

    # Apply all replacements at once
    if replacements:
        components_obj = replace(components_obj, **replacements)

    return components_obj



from dataclasses import dataclass, field
from typing import Any, TypeAlias, get_args

from ruamel import yaml
from ruamel.yaml.comments import CommentedSeq

from ..context import Context
from ..extractors import extract_extension_fields
from ..fields import fixed_field, fixed_fields
from ..sources import FieldSource, KeySource, ValueSource, YAMLValue
from .builders import build_field_source
from .discriminator import Discriminator
from .discriminator import build as build_discriminator
from .external_documentation import ExternalDocumentation
from .external_documentation import build as build_external_documentation
from .xml import XML
from .xml import build as build_xml


__all__ = ["Schema", "BooleanJSONSchema", "NestedSchema", "build"]


BooleanJSONSchema: TypeAlias = ValueSource[bool]


# Type alias for nested schema references
# A schema node that can be nested within another schema, representing:
# - Schema: A valid schema object
# - Boolean JSON Schema: True or False
# - ValueSource[str | int | float | None | CommentedSeq]: Invalid/malformed data preserved for validation

NestedSchema: TypeAlias = (
    "Schema | BooleanJSONSchema | ValueSource[str | int | float | None | CommentedSeq]"
)


@dataclass(frozen=True, slots=True)
class Schema:
    """
    Schema Object representation for OpenAPI 3.1.

    In OpenAPI 3.1, the Schema Object is a full JSON Schema 2020-12 vocabulary with OpenAPI extensions.
    This represents a complete JSON Schema with additional OpenAPI-specific fields.

    Attributes:
        root_node: The top-level node representing the entire Schema object in the original source file

        # JSON Schema Core Keywords (2020-12)
        schema: The $schema keyword - URI of the meta-schema
        id: The $id keyword - URI that identifies the schema resource
        ref: The $ref keyword - URI reference to another schema
        anchor: The $anchor keyword - Plain name fragment for identification
        dynamic_ref: The $dynamicRef keyword - Dynamic reference to another schema
        dynamic_anchor: The $dynamicAnchor keyword - Dynamic anchor for identification
        vocabulary: The $vocabulary keyword - Available vocabularies and their usage
        comment: The $comment keyword - Comments for schema authors
        defs: The $defs keyword - Schema definitions for reuse

        # JSON Schema Validation Keywords (2020-12)
        # Validation - Any Type
        type: The type keyword - Value type(s): string, number, integer, boolean, array, object, null
        enum: The enum keyword - Fixed set of allowed values
        const: The const keyword - Single allowed value

        # Validation - Numeric
        multiple_of: A numeric instance is valid only if division by this value results in an integer
        maximum: Upper limit for a numeric instance (inclusive)
        exclusive_maximum: Upper limit for a numeric instance (exclusive)
        minimum: Lower limit for a numeric instance (inclusive)
        exclusive_minimum: Lower limit for a numeric instance (exclusive)

        # Validation - String
        max_length: Maximum length of a string instance
        min_length: Minimum length of a string instance
        pattern: A string instance is valid if the regular expression matches the instance successfully

        # Validation - Array
        max_items: Maximum number of items in an array instance
        min_items: Minimum number of items in an array instance
        unique_items: If true, array items must be unique
        max_contains: Maximum number of items that must match the contains schema
        min_contains: Minimum number of items that must match the contains schema

        # Validation - Object
        max_properties: Maximum number of properties in an object instance
        min_properties: Minimum number of properties in an object instance
        required: List of required property names
        dependent_required: Dependencies between properties (property-based requirements)

        # JSON Schema Applicator Keywords (2020-12)
        # Schema Composition
        all_of: Must be valid against all of the subschemas
        any_of: Must be valid against any of the subschemas
        one_of: Must be valid against exactly one of the subschemas
        not_: Must not be valid against the given schema

        # Conditional Application
        if_: Conditional schema - if this validates, then/else is applied
        then_: Schema to apply if 'if' validates
        else_: Schema to apply if 'if' does not validate

        # Array Applicators
        prefix_items: Array of schemas for validating tuple-like arrays (positional items)
        items: Schema for array items (all items or items after prefix_items)
        contains: Schema that at least one array item must validate against

        # Object Applicators
        properties: Property name to schema mappings
        pattern_properties: Property patterns to schema mappings
        additional_properties: Schema for properties not defined in properties/pattern_properties, or boolean
        property_names: Schema that all property names must validate against

        # Dependent Schemas
        dependent_schemas: Schemas that must validate when specific properties are present

        # Unevaluated Locations
        unevaluated_items: Schema for array items not evaluated by other keywords
        unevaluated_properties: Schema for object properties not evaluated by other keywords

        # JSON Schema Meta-Data Keywords (2020-12)
        title: A title for the schema
        description: A description of the schema. CommonMark syntax MAY be used for rich text representation.
        default: Default value for the schema
        deprecated: Specifies that the schema is deprecated
        read_only: Indicates the value should not be modified (sent in response but not in request)
        write_only: Indicates the value should only be sent in requests (not in responses)
        examples: Array of example values

        # JSON Schema Format Keywords (2020-12)
        format: Additional format hint for the type (e.g., "email", "uuid", "uri", "date-time")

        # JSON Schema Content Keywords (2020-12)
        content_encoding: Content encoding for string instances (e.g., "base64")
        content_media_type: Media type of string instance contents (e.g., "application/json")
        content_schema: Schema for validating the decoded content

        # OpenAPI-specific extensions (not in JSON Schema 2020-12)
        discriminator: Adds support for polymorphism
        xml: Additional metadata for XML representations
        external_docs: Additional external documentation
        example: A single example value (OpenAPI extension, use examples for standard JSON Schema)

        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node

    # Core Keywords
    id: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$id"})
    schema: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$schema"})
    ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$ref"})
    comment: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$comment"})
    defs: FieldSource[dict[KeySource[str], NestedSchema]] | None = fixed_field(
        metadata={"yaml_name": "$defs"}
    )
    anchor: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$anchor"})
    dynamic_anchor: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$dynamicAnchor"})
    dynamic_ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$dynamicRef"})
    vocabulary: FieldSource[dict[KeySource[str], ValueSource[bool]]] | None = fixed_field(
        metadata={"yaml_name": "$vocabulary"}
    )

    # JSON Schema Applicator Keywords
    all_of: FieldSource[list[NestedSchema]] | None = fixed_field(metadata={"yaml_name": "allOf"})
    any_of: FieldSource[list[NestedSchema]] | None = fixed_field(metadata={"yaml_name": "anyOf"})
    one_of: FieldSource[list[NestedSchema]] | None = fixed_field(metadata={"yaml_name": "oneOf"})
    if_: FieldSource[NestedSchema] | None = fixed_field(metadata={"yaml_name": "if"})
    then_: FieldSource[NestedSchema] | None = fixed_field(metadata={"yaml_name": "then"})
    else_: FieldSource[NestedSchema] | None = fixed_field(metadata={"yaml_name": "else"})
    not_: FieldSource[NestedSchema] | None = fixed_field(metadata={"yaml_name": "not"})
    properties: FieldSource[dict[KeySource[str], NestedSchema]] | None = fixed_field()
    additional_properties: FieldSource[NestedSchema] | None = fixed_field(
        metadata={"yaml_name": "additionalProperties"}
    )
    pattern_properties: FieldSource[dict[KeySource[str], NestedSchema]] | None = fixed_field(
        metadata={"yaml_name": "patternProperties"}
    )
    dependent_schemas: FieldSource[dict[KeySource[str], NestedSchema]] | None = fixed_field(
        metadata={"yaml_name": "dependentSchemas"}
    )
    property_names: FieldSource[NestedSchema] | None = fixed_field(
        metadata={"yaml_name": "propertyNames"}
    )
    contains: FieldSource[NestedSchema] | None = fixed_field()
    items: FieldSource[NestedSchema] | None = fixed_field()
    prefix_items: FieldSource[list[NestedSchema]] | None = fixed_field(
        metadata={"yaml_name": "prefixItems"}
    )

    # Validation Keywords
    type: FieldSource[str | list[ValueSource[str]]] | None = fixed_field()
    enum: FieldSource[list[ValueSource[YAMLValue]]] | None = fixed_field()
    const: FieldSource[YAMLValue] | None = fixed_field()
    max_length: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "maxLength"})
    min_length: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "minLength"})
    pattern: FieldSource[str] | None = fixed_field()
    exclusive_maximum: FieldSource[int | float] | None = fixed_field(
        metadata={"yaml_name": "exclusiveMaximum"}
    )
    exclusive_minimum: FieldSource[int | float] | None = fixed_field(
        metadata={"yaml_name": "exclusiveMinimum"}
    )
    minimum: FieldSource[int | float] | None = fixed_field()
    maximum: FieldSource[int | float] | None = fixed_field()
    multiple_of: FieldSource[int | float] | None = fixed_field(metadata={"yaml_name": "multipleOf"})
    dependent_required: FieldSource[dict[KeySource[str], list[ValueSource[str]]]] | None = (
        fixed_field(metadata={"yaml_name": "dependentRequired"})
    )
    max_properties: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "maxProperties"})
    min_properties: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "minProperties"})
    required: FieldSource[list[ValueSource[str]]] | None = fixed_field()
    max_items: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "maxItems"})
    min_items: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "minItems"})
    max_contains: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "maxContains"})
    min_contains: FieldSource[int] | None = fixed_field(metadata={"yaml_name": "minContains"})
    unique_items: FieldSource[bool] | None = fixed_field(metadata={"yaml_name": "uniqueItems"})

    # Meta Data Keywords
    title: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    default: FieldSource[YAMLValue] | None = fixed_field()
    deprecated: FieldSource[bool] | None = fixed_field()
    examples: FieldSource[list[ValueSource[YAMLValue]]] | None = fixed_field()
    read_only: FieldSource[bool] | None = fixed_field(metadata={"yaml_name": "readOnly"})
    write_only: FieldSource[bool] | None = fixed_field(metadata={"yaml_name": "writeOnly"})

    # Format Annotation Keywords
    format: FieldSource[str] | None = fixed_field()

    # Content Keywords
    content_encoding: FieldSource[str] | None = fixed_field(
        metadata={"yaml_name": "contentEncoding"}
    )
    content_media_type: FieldSource[str] | None = fixed_field(
        metadata={"yaml_name": "contentMediaType"}
    )
    content_schema: FieldSource[NestedSchema] | None = fixed_field(
        metadata={"yaml_name": "contentSchema"}
    )

    # Unevaluated Keywords
    unevaluated_items: FieldSource[NestedSchema] | None = fixed_field(
        metadata={"yaml_name": "unevaluatedItems"}
    )
    unevaluated_properties: FieldSource[NestedSchema] | None = fixed_field(
        metadata={"yaml_name": "unevaluatedProperties"}
    )

    # OpenAPI base vocabulary Keywords (not in JSON Schema 2020-12)
    discriminator: FieldSource[Discriminator] | None = fixed_field()
    xml: FieldSource[XML] | None = fixed_field()
    external_docs: FieldSource[ExternalDocumentation] | None = fixed_field(
        metadata={"yaml_name": "externalDocs"}
    )
    example: FieldSource[YAMLValue] | None = fixed_field()

    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> "Schema | BooleanJSONSchema | ValueSource[str | int | float | None | CommentedSeq]":
    """
    Build a Schema object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Note: Schema is self-referential (can contain other Schema objects in all_of, one_of, any_of, not_,
    items, properties, additional_properties). The builder handles nested Schema objects by preserving
    them as raw YAML values, letting validation layers interpret them.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Schema object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("type: string\\nminLength: 1\\nmaxLength: 100")
        schema = build(root)
        assert schema.type.value == 'string'
        assert schema.min_length.value == 1
    """
    context = context or Context()

    if not isinstance(root, yaml.MappingNode):
        # Preserve invalid root data instead of returning None
        value = context.yaml_constructor.construct_object(root, deep=True)
        return ValueSource(value=value, value_node=root)

    # Build YAML name to Python field name mapping
    _fixed_fields = fixed_fields(Schema)
    yaml_to_field = {
        f.metadata.get("yaml_name", fname): fname for fname, f in _fixed_fields.items()
    }

    # Accumulate all field values in a single pass
    field_values: dict[str, Any] = {}

    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        # Skip extension fields - handled separately at the end
        if key.startswith("x-"):
            continue

        # Map YAML key to Python field name
        field_name = yaml_to_field.get(key)
        if not field_name:
            continue

        # Get field metadata
        field_info = _fixed_fields[field_name]
        field_type_args = set(get_args(field_info.type))

        # Simple scalar fields (handled like build_model does)
        if field_type_args & {
            FieldSource[str],
            FieldSource[bool],
            FieldSource[int],
            FieldSource[int | float],
            FieldSource[YAMLValue],
        }:
            field_values[field_name] = build_field_source(key_node, value_node, context)

        # Handle list with ValueSource wrapping for each item (e.g., required, enum fields)
        elif field_type_args & {
            FieldSource[list[ValueSource[str]]],
            FieldSource[list[ValueSource[YAMLValue]]],
        }:
            if isinstance(value_node, yaml.SequenceNode):
                value_list: list[ValueSource[Any]] = []
                for item_node in value_node.value:
                    item_value = context.yaml_constructor.construct_object(item_node, deep=True)
                    value_list.append(ValueSource(value=item_value, value_node=item_node))
                field_values[field_name] = FieldSource(
                    value=value_list, key_node=key_node, value_node=value_node
                )
            else:
                # Not a sequence - preserve as-is for validation
                field_values[field_name] = build_field_source(key_node, value_node, context)

        # Recursive schema list fields (allOf, oneOf, anyOf, prefixItems)
        elif key in ("allOf", "oneOf", "anyOf", "prefixItems"):
            if isinstance(value_node, yaml.SequenceNode):
                schemas = []
                for item_node in value_node.value:
                    schema = build(item_node, context)
                    schemas.append(schema)
                field_values[field_name] = FieldSource(
                    value=schemas, key_node=key_node, value_node=value_node
                )
            else:
                # Not a sequence - preserve as-is for validation
                field_values[field_name] = build_field_source(key_node, value_node, context)
        # Recursive schema single fields (not, if, then, else, contains, propertyNames, contentSchema)
        elif key in ("not", "if", "then", "else", "contains", "propertyNames", "contentSchema"):
            schema = build(value_node, context)
            field_values[field_name] = FieldSource(
                value=schema, key_node=key_node, value_node=value_node
            )
        # items (boolean | schema in JSON Schema 2020-12)
        elif key == "items":
            # Check if it's a boolean or a schema
            if (
                isinstance(value_node, yaml.ScalarNode)
                and value_node.tag == "tag:yaml.org,2002:bool"
            ):
                field_values[field_name] = build_field_source(key_node, value_node, context)
            else:
                # It's a schema
                schema = build(value_node, context)
                field_values[field_name] = FieldSource(
                    value=schema, key_node=key_node, value_node=value_node
                )
        # Boolean | schema fields (additionalProperties, unevaluatedItems, unevaluatedProperties)
        elif key in ("additionalProperties", "unevaluatedItems", "unevaluatedProperties"):
            # Check if it's a boolean or a schema
            if (
                isinstance(value_node, yaml.ScalarNode)
                and value_node.tag == "tag:yaml.org,2002:bool"
            ):
                field_values[field_name] = build_field_source(key_node, value_node, context)
            else:
                # It's a schema
                schema = build(value_node, context)
                field_values[field_name] = FieldSource(
                    value=schema, key_node=key_node, value_node=value_node
                )
        # Dict of schemas (properties, $defs, patternProperties, dependentSchemas)
        elif key in ("properties", "$defs", "patternProperties", "dependentSchemas"):
            if isinstance(value_node, yaml.MappingNode):
                schemas_dict: dict[KeySource[str], NestedSchema] = {}
                for map_key_node, map_value_node in value_node.value:
                    map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                    # Recursively build each schema
                    nested_schema: NestedSchema = build(map_value_node, context)
                    schemas_dict[KeySource(value=map_key, key_node=map_key_node)] = nested_schema
                field_values[field_name] = FieldSource(
                    value=schemas_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                field_values[field_name] = build_field_source(key_node, value_node, context)
        # $vocabulary (dict[KeySource[str], ValueSource[bool]])
        elif key == "$vocabulary":
            if isinstance(value_node, yaml.MappingNode):
                vocabulary_dict: dict[KeySource[str], ValueSource[bool]] = {}
                for vocab_key_node, vocab_value_node in value_node.value:
                    vocab_key = context.yaml_constructor.construct_yaml_str(vocab_key_node)
                    vocab_value = context.yaml_constructor.construct_object(
                        vocab_value_node, deep=True
                    )
                    vocabulary_dict[KeySource(value=vocab_key, key_node=vocab_key_node)] = (
                        ValueSource(value=vocab_value, value_node=vocab_value_node)
                    )
                field_values[field_name] = FieldSource(
                    value=vocabulary_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                field_values[field_name] = build_field_source(key_node, value_node, context)
        # dependentRequired (dict[KeySource[str], list[ValueSource[str]]])
        elif key == "dependentRequired":
            if isinstance(value_node, yaml.MappingNode):
                dependent_required_dict: dict[KeySource[str], list[ValueSource[str]]] = {}
                for dep_key_node, dep_value_node in value_node.value:
                    dep_key = context.yaml_constructor.construct_yaml_str(dep_key_node)
                    if isinstance(dep_value_node, yaml.SequenceNode):
                        dep_list: list[ValueSource[str]] = []
                        for item_node in dep_value_node.value:
                            item_value = context.yaml_constructor.construct_object(
                                item_node, deep=True
                            )
                            dep_list.append(ValueSource(value=item_value, value_node=item_node))
                        dependent_required_dict[KeySource(value=dep_key, key_node=dep_key_node)] = (
                            dep_list
                        )
                    else:
                        # Not a sequence - preserve as invalid for validation
                        invalid_value = context.yaml_constructor.construct_object(
                            dep_value_node, deep=True
                        )
                        dependent_required_dict[KeySource(value=dep_key, key_node=dep_key_node)] = (
                            invalid_value  # type: ignore
                        )
                field_values[field_name] = FieldSource(
                    value=dependent_required_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                field_values[field_name] = build_field_source(key_node, value_node, context)
        # type field (can be string or list of strings in JSON Schema 2020-12)
        elif key == "type":
            if isinstance(value_node, yaml.SequenceNode):
                # It's a list of types
                type_list: list[ValueSource[str]] = []
                for item_node in value_node.value:
                    item_value = context.yaml_constructor.construct_object(item_node, deep=True)
                    type_list.append(ValueSource(value=item_value, value_node=item_node))
                field_values[field_name] = FieldSource(
                    value=type_list, key_node=key_node, value_node=value_node
                )
            else:
                # It's a single type string
                field_values[field_name] = build_field_source(key_node, value_node, context)
        # Nested objects (discriminator, xml, externalDocs)
        elif key == "discriminator":
            field_values[field_name] = FieldSource(
                value=build_discriminator(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
        elif key == "xml":
            field_values[field_name] = FieldSource(
                value=build_xml(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
        elif key == "externalDocs":
            field_values[field_name] = FieldSource(
                value=build_external_documentation(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )

    # Build and return the Schema instance (single constructor call)
    return Schema(
        root_node=root,
        **field_values,
        extensions=extract_extension_fields(root, context),
    )



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model


__all__ = ["Discriminator", "build"]


@dataclass(frozen=True, slots=True)
class Discriminator:
    """
    Discriminator Object representation for OpenAPI 3.1.

    When request bodies or response payloads may be one of a number of different schemas,
    a discriminator object can be used to aid in serialization, deserialization, and validation.

    In OpenAPI 3.1, the Discriminator Object supports Specification Extensions (x-* fields).

    Attributes:
        root_node: The top-level node representing the entire Discriminator object in the original source file
        property_name: REQUIRED. The name of the property in the payload that will hold the discriminator value
        mapping: An optional mapping of discriminator values to schema names or references
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    property_name: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "propertyName"})
    mapping: FieldSource[dict[KeySource[str], ValueSource[str]]] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Discriminator | ValueSource[YAMLInvalidValue]:
    """
    Build a Discriminator object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Discriminator object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("propertyName: petType\\nmapping:\\n  dog: Dog\\n  cat: Cat")
        discriminator = build(root)
        assert discriminator.property_name.value == 'petType'
    """
    return build_model(root, Discriminator, context=context)



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model


__all__ = ["ExternalDocumentation", "build"]


@dataclass(frozen=True, slots=True)
class ExternalDocumentation:
    """
    External Documentation Object representation for OpenAPI 3.0.

    Allows referencing an external resource for extended documentation.

    Attributes:
        root_node: The top-level node representing the entire External Documentation object in the original source file
        description: A short description of the target documentation. CommonMark syntax MAY be used for rich text representation.
        url: The URL for the target documentation. REQUIRED. Value MUST be in the format of a URL.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    description: FieldSource[str] | None = fixed_field()
    url: FieldSource[str] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> ExternalDocumentation | ValueSource[YAMLInvalidValue]:
    """
    Build an ExternalDocumentation object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An ExternalDocumentation object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("url: https://example.com\\ndescription: Find more info here")
        external_docs = build(root)
        assert external_docs.url.value == 'https://example.com'
    """
    return build_model(root, ExternalDocumentation, context=context)



from dataclasses import dataclass
from typing import Any

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, ValueSource, YAMLInvalidValue
from .builders import build_model


__all__ = ["Reference", "build"]


@dataclass(frozen=True, slots=True)
class Reference:
    """
    Reference Object representation for OpenAPI 3.1.

    Allows for a reference to another document.

    Attributes:
        root_node: The top-level node representing the entire Reference object in the original source file
        meta: Optional metadata as a dictionary.
        ref: REQUIRED. The reference identifier. This MUST be in the form of a URI.
        summary: A short summary which by default SHOULD override that of the referenced component. If the referenced object-type does not allow a summary field, then this field has no effect.
        description: A description which by default SHOULD override that of the referenced component. CommonMark syntax MAY be used for rich text representation. If the referenced object-type does not allow a description field, then this field has no effect.
    """

    root_node: yaml.Node
    meta: dict[str, Any] | None = None
    ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$ref"})
    summary: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()


def build(
    root: yaml.Node, context: Context | None = None
) -> Reference | ValueSource[YAMLInvalidValue]:
    """
    Build a Reference object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Reference object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        $ref: '#/components/schemas/Pet'
        summary: A Pet reference
        description: Reference to the Pet schema in components
        ''')
        reference = build(root)
        assert reference.ref.value == '#/components/schemas/Pet'
        assert reference.summary.value == 'A Pet reference'
        assert reference.description.value == 'Reference to the Pet schema in components'
    """
    return build_model(root, Reference, context=context)



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_field_source, build_model
from .server_variable import ServerVariable
from .server_variable import build as build_server_variable


__all__ = ["Server", "build"]


@dataclass(frozen=True, slots=True)
class Server:
    """
    Server Object representation for OpenAPI 3.0.

    An object representing a Server.

    Attributes:
        root_node: The top-level node representing the entire Server object in the original source file
        url: REQUIRED. A URL to the target host. This URL supports Server Variables and MAY be relative, to indicate that the host location is relative to the location where the OpenAPI document is being served. Variable substitutions will be made when a variable is named in {brackets}.
        description: An optional string describing the host designated by the URL. CommonMark syntax MAY be used for rich text representation.
        variables: A map between a variable name and its value. The value is used for substitution in the server's URL template.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    url: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    variables: FieldSource[dict[KeySource[str], ServerVariable]] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Server | ValueSource[YAMLInvalidValue]:
    """
    Build a Server object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Server object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        url: https://{environment}.example.com/api/v1
        description: Production API server
        variables:
          environment:
            default: production
            enum:
              - production
              - staging
              - development
            description: The deployment environment
        ''')
        server = build(root)
        assert server.url.value == 'https://{environment}.example.com/api/v1'
        assert server.description.value == 'Production API server'
        assert 'environment' in {k.value for k in server.variables.value.keys()}
    """
    context = context or Context()

    # Use build_model for initial construction
    server = build_model(root, Server, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(server, Server):
        return server

    # Manually handle nested variables field
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "variables":
            # Handle variables field - map of ServerVariable objects
            if isinstance(value_node, yaml.MappingNode):
                variables_dict: dict[
                    KeySource[str], ServerVariable | ValueSource[YAMLInvalidValue]
                ] = {}
                for var_key_node, var_value_node in value_node.value:
                    var_key = context.yaml_constructor.construct_yaml_str(var_key_node)
                    # Build ServerVariable for each variable - child builder handles invalid nodes
                    variables_dict[KeySource(value=var_key, key_node=var_key_node)] = (
                        build_server_variable(var_value_node, context=context)
                    )
                variables = FieldSource(
                    value=variables_dict, key_node=key_node, value_node=value_node
                )
                server = replace(server, variables=variables)
            else:
                # Not a mapping - preserve as-is for validation
                variables = build_field_source(key_node, value_node, context)
                server = replace(server, variables=variables)
            break

    return server



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_field_source, build_model
from .reference import Reference
from .reference import build as build_reference
from .server import Server
from .server import build as build_server


__all__ = ["Link", "build", "build_link_or_reference"]


@dataclass(frozen=True, slots=True)
class Link:
    """
    Link Object representation for OpenAPI 3.0.

    The Link Object represents a possible design-time link for a response.

    Attributes:
        root_node: The top-level node representing the entire Link object in the original source file
        operation_ref: A relative or absolute URI reference to an OAS operation.
        operation_id: The name of an existing, resolvable OAS operation.
        parameters: A map representing parameters to pass to an operation as specified with operationId
                   or identified via operationRef.
        request_body: A literal value or {expression} to use as a request body when calling the target operation.
        description: A description of the link. CommonMark syntax MAY be used for rich text representation.
        server: A server object to be used by the target operation.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    operation_ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "operationRef"})
    operation_id: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "operationId"})
    parameters: FieldSource[dict[KeySource[str], ValueSource[YAMLValue]]] | None = fixed_field()
    request_body: FieldSource[YAMLValue] | None = fixed_field(metadata={"yaml_name": "requestBody"})
    description: FieldSource[str] | None = fixed_field()
    server: FieldSource[Server] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(root: yaml.Node, context: Context | None = None) -> Link | ValueSource[YAMLInvalidValue]:
    """
    Build a Link object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Link object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("operationId: getUserById\\nparameters:\\n  userId: $response.body#/id")
        link = build(root)
        assert link.operation_id.value == 'getUserById'
    """
    context = context or Context()

    # Use build_model for initial construction
    link = build_model(root, Link, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(link, Link):
        return link

    # Manually handle nested server object and parameters dict
    replacements = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "server":
            # Handle nested Server object - child builder handles invalid nodes
            replacements["server"] = FieldSource(
                value=build_server(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
        elif key == "parameters":
            # Handle parameters map with KeySource/ValueSource wrapping
            if isinstance(value_node, yaml.MappingNode):
                params_dict: dict[KeySource[str], ValueSource[YAMLValue]] = {}
                for param_key_node, param_value_node in value_node.value:
                    param_key = context.yaml_constructor.construct_yaml_str(param_key_node)
                    param_value = context.yaml_constructor.construct_object(
                        param_value_node, deep=True
                    )
                    params_dict[KeySource(value=param_key, key_node=param_key_node)] = ValueSource(
                        value=param_value, value_node=param_value_node
                    )
                replacements["parameters"] = FieldSource(
                    value=params_dict, key_node=key_node, value_node=value_node
                )
            else:
                # Not a mapping - preserve as-is for validation
                replacements["parameters"] = build_field_source(key_node, value_node, context)

    # Apply all replacements at once
    if replacements:
        link = replace(link, **replacements)

    return link


def build_link_or_reference(
    node: yaml.Node, context: Context
) -> Link | Reference | ValueSource[YAMLInvalidValue]:
    """
    Build either a Link or Reference from a YAML node.

    This helper handles the polymorphic nature of OpenAPI where many fields
    can contain either a Link object or a Reference object ($ref).

    Args:
        node: The YAML node to parse
        context: Parsing context

    Returns:
        A Link, Reference, or ValueSource if the node is invalid
    """
    # Check if it's a reference (has $ref key)
    if isinstance(node, yaml.MappingNode):
        for key_node, _ in node.value:
            key = context.yaml_constructor.construct_yaml_str(key_node)
            if key == "$ref":
                return build_reference(node, context)

    # Otherwise, try to build as Link
    return build(node, context)



from dataclasses import dataclass
from typing import Any

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, ValueSource, YAMLInvalidValue
from .builders import build_model


__all__ = ["Reference", "build"]


@dataclass(frozen=True, slots=True)
class Reference:
    """
    Reference Object representation for OpenAPI 3.0.

    A simple object to allow referencing other components in the OpenAPI Description,
    internally and externally.

    Note: In OpenAPI 3.0, Reference Objects only have the $ref field.
    Summary, description, and extensions were added in OpenAPI 3.1.

    Attributes:
        root_node: The top-level node representing the entire Reference object in the original source file
        meta: Optional metadata as a dictionary.
        ref: REQUIRED. The reference string. Must be in the format of a URI.
    """

    root_node: yaml.Node
    meta: dict[str, Any] | None = None
    ref: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "$ref"})


def build(
    root: yaml.Node, context: Context | None = None
) -> Reference | ValueSource[YAMLInvalidValue]:
    """
    Build a Reference object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Reference object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("$ref: '#/components/schemas/Pet'")
        reference = build(root)
        assert reference.ref.value == '#/components/schemas/Pet'
    """
    return build_model(root, Reference, context=context)



import re
from dataclasses import dataclass, field
from typing import cast

from ruamel import yaml

from ..context import Context
from ..extractors import extract_extension_fields
from ..fields import fixed_field, patterned_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .reference import Reference
from .response import Response, build_response_or_reference


__all__ = ["Responses", "build"]


@dataclass(frozen=True, slots=True)
class Responses:
    """
    Responses Object representation for OpenAPI 3.1.

    A container for the expected responses of an operation. The container maps a HTTP response code
    to the expected response.

    Attributes:
        root_node: The top-level node representing the entire Responses object in the original source file
        default: The documentation of responses other than the ones declared for specific HTTP response codes.
                Use this field to cover undeclared responses.
        responses: Maps HTTP status codes to their Response objects. The key is the HTTP status code
                  (e.g., "200", "404", "4XX") and the value is a Response object or Reference.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    default: FieldSource[Response | Reference] | None = fixed_field()
    responses: dict[KeySource[str], Response | Reference] = patterned_field(default_factory=dict)
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Responses | ValueSource[YAMLInvalidValue]:
    """
    Build a Responses object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A Responses object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        '200':
          description: successful operation
        '404':
          description: not found
        default:
          description: unexpected error
        ''')
        responses = build(root)
        assert '200' in {k.value for k in responses.responses.keys()}
    """
    context = context or Context()

    # Check if root is a MappingNode, if not return ValueSource with invalid data
    if not isinstance(root, yaml.MappingNode):
        value = context.yaml_constructor.construct_object(root, deep=True)
        return ValueSource(value=value, value_node=root)

    # Process each field to determine if it's default or a status code response
    responses = {}
    default: FieldSource[Response | Reference] | None = None

    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "default":
            # Handle default response - can be Response or Reference
            response_or_reference = build_response_or_reference(value_node, context)
            default = cast(
                FieldSource[Response | Reference],
                FieldSource(value=response_or_reference, key_node=key_node, value_node=value_node),
            )
        elif _HTTP_STATUS_CODE_PATTERN.match(key):
            # Valid HTTP status code (100-599) or pattern (1XX-5XX)
            response_or_reference = build_response_or_reference(value_node, context)
            responses[KeySource(value=key, key_node=key_node)] = response_or_reference

    # Create and return the Responses object with collected data
    return Responses(
        root_node=root,
        default=default,
        responses=responses,
        extensions=extract_extension_fields(root, context),
    )


# Pattern for valid HTTP status codes: 100-599 or wildcard patterns (1XX-5XX)
_HTTP_STATUS_CODE_PATTERN = re.compile(r"^([1-5]XX|[1-5][0-9]{2})$")



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .contact import Contact
from .contact import build as build_contact
from .license import License
from .license import build as build_license


__all__ = ["Info", "build"]


@dataclass(frozen=True, slots=True)
class Info:
    """
    Info Object representation for OpenAPI 3.1.

    Provides metadata about the API. The metadata MAY be used by the clients if needed,
    and MAY be presented in editing or documentation generation tools for convenience.

    Attributes:
        root_node: The top-level node representing the entire Info object in the original source file
        title: REQUIRED. The title of the API.
        summary: A short summary of the API (new in OpenAPI 3.1).
        description: A description of the API. CommonMark syntax MAY be used for rich text representation.
        termsOfService: A URL to the Terms of Service for the API. This MUST be in the form of a URL.
        contact: The contact information for the exposed API.
        license: The license information for the exposed API.
        version: REQUIRED. The version of the OpenAPI document (which is distinct from the OpenAPI Specification version or the API implementation version).
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    title: FieldSource[str] | None = fixed_field()
    summary: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    termsOfService: FieldSource[str] | None = fixed_field()
    contact: FieldSource[Contact] | None = fixed_field()
    license: FieldSource[License] | None = fixed_field()
    version: FieldSource[str] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(root: yaml.Node, context: Context | None = None) -> Info | ValueSource[YAMLInvalidValue]:
    """
    Build an Info object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An Info object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        title: Sample Pet Store App
        version: 1.0.1
        description: This is a sample server for a pet store.
        contact:
          name: API Support
          email: support@example.com
        license:
          name: Apache 2.0
          url: https://www.apache.org/licenses/LICENSE-2.0.html
        ''')
        info = build(root)
        assert info.title.value == 'Sample Pet Store App'
        assert info.version.value == '1.0.1'
        assert info.contact.value.name.value == 'API Support'
    """
    context = context or Context()

    # Use build_model for initial construction
    info = build_model(root, Info, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(info, Info):
        return info

    # Manually handle nested objects (contact, license)
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "contact":
            # Handle nested Contact object - child builder handles invalid nodes
            # FieldSource will auto-unwrap ValueSource if child returns it for invalid data
            contact = FieldSource(
                value=build_contact(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
            info = replace(info, contact=contact)
        elif key == "license":
            # Handle nested License object - child builder handles invalid nodes
            # FieldSource will auto-unwrap ValueSource if child returns it for invalid data
            license_obj = FieldSource(
                value=build_license(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
            info = replace(info, license=license_obj)

    return info



"""Builders for OpenAPI 3.1.x specification objects."""

from dataclasses import fields
from typing import TYPE_CHECKING, Any, TypeVar, cast, get_args

from ruamel import yaml

from ...context import Context
from ...extractors import extract_extension_fields
from ...fields import fixed_fields
from ...sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue


if TYPE_CHECKING:
    from ..callback import Callback  # noqa: F401
    from ..example import Example  # noqa: F401
    from ..external_documentation import ExternalDocumentation  # noqa: F401
    from ..header import Header  # noqa: F401
    from ..link import Link  # noqa: F401
    from ..media_type import MediaType  # noqa: F401
    from ..parameter import Parameter  # noqa: F401
    from ..path_item import PathItem  # noqa: F401
    from ..reference import Reference  # noqa: F401
    from ..schema import BooleanJSONSchema, Schema  # noqa: F401
    from ..security_requirement import SecurityRequirement  # noqa: F401
    from ..server import Server  # noqa: F401


__all__ = ["build_model", "build_field_source"]


T = TypeVar("T")


def build_model(
    root: yaml.Node, dataclass_type: type[T], *, context: Context | None = None
) -> T | ValueSource[YAMLInvalidValue]:
    """
    Generic builder for OpenAPI 3.1 low model.

    Builds any dataclass that follows the pattern:
    - Has a required `root_node: yaml.Node` field
    - Has an optional `extensions: dict[...]` field
    - Has spec fields marked with `fixed_field()`

    Args:
        root: The YAML node to parse (should be a MappingNode)
        dataclass_type: The dataclass type to build
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An instance of dataclass_type if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        xml = build_model(root_node, XML, context=context)
    """
    # Initialize context once at the beginning
    if context is None:
        context = Context()

    if not isinstance(root, yaml.MappingNode):
        # Preserve invalid root data instead of returning None
        value = context.yaml_constructor.construct_object(root, deep=True)
        return ValueSource(value=value, value_node=root)

    # Get fixed specification fields for this dataclass type
    _fixed_fields = fixed_fields(dataclass_type)

    # Build YAML name to Python field name mapping
    yaml_to_field = {
        field.metadata.get("yaml_name", fname): fname for fname, field in _fixed_fields.items()
    }

    # Extract field values in a single pass (non-recursive, single layer only)
    field_values: dict[str, FieldSource[Any]] = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        # Map YAML key to Python field name
        field_name = yaml_to_field.get(key)
        if field_name:
            field = _fixed_fields[field_name]
            field_type_args = set(get_args(field.type))

            if field_type_args & {
                FieldSource[str],
                FieldSource[bool],
                FieldSource[int],
                FieldSource[YAMLValue],
            }:
                field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], ValueSource[str]]]}:
                # Handle dict with KeySource/ValueSource wrapping
                if isinstance(value_node, yaml.MappingNode):
                    mapping_dict: dict[KeySource[str], ValueSource[str]] = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        map_value = context.yaml_constructor.construct_object(
                            map_value_node, deep=True
                        )
                        mapping_dict[KeySource(value=map_key, key_node=map_key_node)] = ValueSource(
                            value=map_value, value_node=map_value_node
                        )
                    field_values[field_name] = FieldSource(
                        value=mapping_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[list[ValueSource[str]]]}:
                # Handle list with ValueSource wrapping for each item
                if isinstance(value_node, yaml.SequenceNode):
                    value_list: list[ValueSource[str]] = []
                    for item_node in value_node.value:
                        item_value = context.yaml_constructor.construct_object(item_node, deep=True)
                        value_list.append(ValueSource(value=item_value, value_node=item_node))
                    field_values[field_name] = FieldSource(
                        value=value_list, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a sequence - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[list["Server"]]}:
                # Handle list[Server] with lazy import
                from ..server import build as build_server

                if isinstance(value_node, yaml.SequenceNode):
                    servers_list = []
                    for item_node in value_node.value:
                        server_obj = build_server(item_node, context)
                        servers_list.append(server_obj)
                    field_values[field_name] = FieldSource(
                        value=servers_list, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a sequence - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[list["SecurityRequirement"]]}:
                # Handle list[SecurityRequirement] with lazy import
                from ..security_requirement import build as build_security_requirement

                if isinstance(value_node, yaml.SequenceNode):
                    security_list = []
                    for item_node in value_node.value:
                        security_req = build_security_requirement(item_node, context)
                        security_list.append(security_req)
                    field_values[field_name] = FieldSource(
                        value=security_list, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a sequence - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[list["Parameter | Reference"]]}:
                # Handle list[Parameter | Reference] with lazy import
                from ..parameter import build_parameter_or_reference

                if isinstance(value_node, yaml.SequenceNode):
                    parameters_list = []
                    for item_node in value_node.value:
                        parameter_or_reference = build_parameter_or_reference(item_node, context)
                        parameters_list.append(parameter_or_reference)
                    field_values[field_name] = FieldSource(
                        value=parameters_list, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a sequence - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "Example | Reference"]]}:
                # Handle dict[KeySource[str], Example | Reference] with lazy import
                from ..example import build_example_or_reference

                if isinstance(value_node, yaml.MappingNode):
                    examples_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        example_or_reference = build_example_or_reference(map_value_node, context)
                        examples_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            example_or_reference
                        )
                    field_values[field_name] = FieldSource(
                        value=examples_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "MediaType"]]}:
                # Handle dict[KeySource[str], MediaType] with lazy import
                from ..media_type import build as build_media_type

                if isinstance(value_node, yaml.MappingNode):
                    content_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        media_type_obj = build_media_type(map_value_node, context)
                        content_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            media_type_obj
                        )
                    field_values[field_name] = FieldSource(
                        value=content_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "Header | Reference"]]}:
                # Handle dict[KeySource[str], Header | Reference] with lazy import
                from ..header import build_header_or_reference

                if isinstance(value_node, yaml.MappingNode):
                    headers_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        header_or_reference = build_header_or_reference(map_value_node, context)
                        headers_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            header_or_reference
                        )
                    field_values[field_name] = FieldSource(
                        value=headers_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource["ExternalDocumentation"]}:
                # Handle ExternalDocumentation with lazy import
                from ..external_documentation import build as build_external_docs

                field_values[field_name] = FieldSource(
                    value=build_external_docs(value_node, context),
                    key_node=key_node,
                    value_node=value_node,
                )
            elif field_type_args & {FieldSource["Schema | BooleanJSONSchema"]}:
                # Handle Schema | Reference union with lazy import
                from ..schema import build as build_schema

                field_values[field_name] = FieldSource(
                    value=build_schema(value_node, context),
                    key_node=key_node,
                    value_node=value_node,
                )
            elif field_type_args & {FieldSource[dict[KeySource[str], "PathItem"]]}:
                # Handle dict[KeySource[str], PathItem] with lazy import
                from ..path_item import build as build_path_item

                if isinstance(value_node, yaml.MappingNode):
                    path_items_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        path_item = build_path_item(map_value_node, context)
                        path_items_dict[KeySource(value=map_key, key_node=map_key_node)] = path_item
                    field_values[field_name] = FieldSource(
                        value=path_items_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "Callback | Reference"]]}:
                # Handle dict[KeySource[str], Callback | Reference] with lazy import
                from ..callback import build_callback_or_reference

                if isinstance(value_node, yaml.MappingNode):
                    callbacks_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        callback_or_reference = build_callback_or_reference(map_value_node, context)
                        callbacks_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            callback_or_reference
                        )
                    field_values[field_name] = FieldSource(
                        value=callbacks_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "Header | Reference"]]}:
                # Handle dict[KeySource[str], Header | Reference] with lazy import
                from ..header import build_header_or_reference

                if isinstance(value_node, yaml.MappingNode):
                    headers_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        header_or_reference = build_header_or_reference(map_value_node, context)
                        headers_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            header_or_reference
                        )
                    field_values[field_name] = FieldSource(
                        value=headers_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)
            elif field_type_args & {FieldSource[dict[KeySource[str], "Link | Reference"]]}:
                # Handle dict[KeySource[str], Link | Reference] with lazy import
                from ..link import build_link_or_reference

                if isinstance(value_node, yaml.MappingNode):
                    links_dict = {}
                    for map_key_node, map_value_node in value_node.value:
                        map_key = context.yaml_constructor.construct_yaml_str(map_key_node)
                        link_or_reference = build_link_or_reference(map_value_node, context)
                        links_dict[KeySource(value=map_key, key_node=map_key_node)] = (
                            link_or_reference
                        )
                    field_values[field_name] = FieldSource(
                        value=links_dict, key_node=key_node, value_node=value_node
                    )
                else:
                    # Not a mapping - preserve as-is for validation
                    field_values[field_name] = build_field_source(key_node, value_node, context)

    # Build and return the dataclass instance
    # Conditionally include extensions field if dataclass supports it
    # Cast to Any to work around generic type constraints
    has_extensions = any(f.name == "extensions" for f in fields(cast(Any, dataclass_type)))
    return cast(
        T,
        dataclass_type(
            root_node=root,  # type: ignore[call-arg]
            **field_values,
            **({"extensions": extract_extension_fields(root, context)} if has_extensions else {}),
        ),
    )


def build_field_source(
    key_node: yaml.Node,
    value_node: yaml.Node,
    context: Context,
) -> FieldSource[Any]:
    """
    Build a FieldSource from a YAML node.

    Constructs the Python value from the YAML node and wraps it in a FieldSource
    that preserves the original node locations for error reporting.

    Args:
        key_node: The YAML node for the field key
        value_node: The YAML node for the field value
        context: Parsing context

    Returns:
        A FieldSource containing the constructed value and node references
    """
    value = context.yaml_constructor.construct_object(value_node, deep=True)
    return FieldSource(value=value, key_node=key_node, value_node=value_node)



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .oauth_flows import OAuthFlows
from .oauth_flows import build as build_oauth_flows
from .reference import Reference
from .reference import build as build_reference


__all__ = ["SecurityScheme", "build", "build_security_scheme_or_reference"]


@dataclass(frozen=True, slots=True)
class SecurityScheme:
    """
    Security Scheme Object representation for OpenAPI 3.1.

    Defines a security scheme that can be used by operations.
    The security scheme type determines which additional fields are required.

    Attributes:
        root_node: The top-level node representing the entire Security Scheme object in the original source file
        type: REQUIRED. The type of the security scheme. Valid values: "apiKey", "http", "oauth2", "openIdConnect"
        description: A short description for the security scheme. CommonMark syntax MAY be used for rich text representation
        name: REQUIRED for apiKey. The name of the header, query or cookie parameter to be used
        in_: REQUIRED for apiKey. The location of the API key. Valid values: "query", "header", "cookie"
        scheme: REQUIRED for http. The name of the HTTP Authorization scheme (e.g., "basic", "bearer")
        bearer_format: A hint to the client to identify how the bearer token is formatted (e.g., "JWT")
        flows: REQUIRED for oauth2. An object containing configuration information for the flow types supported
        openid_connect_url: REQUIRED for openIdConnect. OpenId Connect URL to discover OAuth2 configuration values
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    type: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    name: FieldSource[str] | None = fixed_field()
    in_: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "in"})
    scheme: FieldSource[str] | None = fixed_field()
    bearer_format: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "bearerFormat"})
    flows: FieldSource[OAuthFlows] | None = fixed_field()
    openid_connect_url: FieldSource[str] | None = fixed_field(
        metadata={"yaml_name": "openIdConnectUrl"}
    )
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> SecurityScheme | ValueSource[YAMLInvalidValue]:
    """
    Build a SecurityScheme object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A SecurityScheme object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("type: apiKey\\nname: api_key\\nin: header")
        security_scheme = build(root)
        assert security_scheme.type.value == 'apiKey'
    """
    context = context or Context()

    # Use build_model for initial construction
    security_scheme = build_model(root, SecurityScheme, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(security_scheme, SecurityScheme):
        return security_scheme

    # Manually handle special fields that build_model can't process (nested objects)
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "flows":
            # Handle nested OAuthFlows object - child builder handles invalid nodes
            # FieldSource will auto-unwrap ValueSource if child returns it for invalid data
            flows = FieldSource(
                value=build_oauth_flows(value_node, context=context),
                key_node=key_node,
                value_node=value_node,
            )
            security_scheme = replace(security_scheme, flows=flows)
            break

    return security_scheme


def build_security_scheme_or_reference(
    node: yaml.Node, context: Context
) -> SecurityScheme | Reference | ValueSource[YAMLInvalidValue]:
    """
    Build either a SecurityScheme or Reference from a YAML node.

    This helper handles the polymorphic nature of OpenAPI where many fields
    can contain either a SecurityScheme object or a Reference object ($ref).

    Args:
        node: The YAML node to parse
        context: Parsing context

    Returns:
        A SecurityScheme, Reference, or ValueSource if the node is invalid
    """
    # Check if it's a reference (has $ref key)
    if isinstance(node, yaml.MappingNode):
        for key_node, _ in node.value:
            key = context.yaml_constructor.construct_yaml_str(key_node)
            if key == "$ref":
                return build_reference(node, context)

    # Otherwise, try to build as SecurityScheme
    return build(node, context)



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model


__all__ = ["ServerVariable", "build"]


@dataclass(frozen=True, slots=True)
class ServerVariable:
    """
    Server Variable Object representation for OpenAPI 3.0.

    An object representing a Server Variable for server URL template substitution.

    Attributes:
        root_node: The top-level node representing the entire ServerVariable object in the original source file
        enum: An enumeration of string values to be used if the substitution options are from a limited set. The array SHOULD NOT be empty.
        default: REQUIRED. The default value to use for substitution, which SHALL be sent if an alternate value is not supplied. If the enum is defined, the value SHOULD exist in the enum's values.
        description: An optional description for the server variable. CommonMark syntax MAY be used for rich text representation.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    enum: FieldSource[list[ValueSource[str]]] | None = fixed_field()
    default: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> ServerVariable | ValueSource[YAMLInvalidValue]:
    """
    Build a ServerVariable object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        A ServerVariable object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        default: production
        enum:
          - production
          - staging
          - development
        description: The deployment environment
        ''')
        server_variable = build(root)
        assert server_variable.default.value == 'production'
        assert len(server_variable.enum.value) == 3
        assert server_variable.enum.value[0].value == 'production'
    """
    return build_model(root, ServerVariable, context=context)



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model


__all__ = ["XML", "build"]


@dataclass(frozen=True, slots=True)
class XML:
    """
    XML Object representation for OpenAPI 3.1.

    Attributes:
        root_node: The top-level node representing the entire XML object in the original source file
        name: Name of the XML element
        namespace: XML namespace
        prefix: XML namespace prefix
        attribute: Whether property is an attribute (deprecated in OpenAPI 3.2+)
        wrapped: Whether array is wrapped
        extensions: Specification extensions
    """

    root_node: yaml.Node
    name: FieldSource[str] | None = fixed_field()
    namespace: FieldSource[str] | None = fixed_field()
    prefix: FieldSource[str] | None = fixed_field()
    attribute: FieldSource[bool] | None = fixed_field()
    wrapped: FieldSource[bool] | None = fixed_field()
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(root: yaml.Node, context: Context | None = None) -> XML | ValueSource[YAMLInvalidValue]:
    """
    Build an XML object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An XML object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).
    """
    return build_model(root, XML, context=context)



from dataclasses import dataclass, field

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_model
from .reference import Reference
from .reference import build as build_reference


__all__ = ["Example", "build", "build_example_or_reference"]


@dataclass(frozen=True, slots=True)
class Example:
    """
    Example Object representation for OpenAPI 3.1.

    Attributes:
        root_node: The top-level node representing the entire Example object in the original source file
        summary: A short description of the example.
        description: A long description for the example. CommonMark syntax MAY be used for rich text representation.
        value: Embedded literal example. The value field and externalValue field are mutually exclusive.
        external_value: A URL that points to the literal example. This provides the capability to reference
                       examples that cannot be inlined. The value field and externalValue field are mutually exclusive.
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    summary: FieldSource[str] | None = fixed_field()
    description: FieldSource[str] | None = fixed_field()
    value: FieldSource[YAMLValue] | None = fixed_field()
    external_value: FieldSource[str] | None = fixed_field(metadata={"yaml_name": "externalValue"})
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> Example | ValueSource[YAMLInvalidValue]:
    """
    Build an Example object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An Example object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose("summary: A user example\\nvalue:\\n  id: 1\\n  name: John")
        example = build(root)
        assert example.summary.value == 'A user example'
    """
    return build_model(root, Example, context=context)


def build_example_or_reference(
    node: yaml.Node, context: Context
) -> Example | Reference | ValueSource[YAMLInvalidValue]:
    """
    Build either an Example or Reference from a YAML node.

    This helper handles the polymorphic nature of OpenAPI where many fields
    can contain either an Example object or a Reference object ($ref).

    Args:
        node: The YAML node to parse
        context: Parsing context

    Returns:
        An Example, Reference, or ValueSource if the node is invalid
    """
    # Check if it's a reference (has $ref key)
    if isinstance(node, yaml.MappingNode):
        for key_node, _ in node.value:
            key = context.yaml_constructor.construct_yaml_str(key_node)
            if key == "$ref":
                return build_reference(node, context)

    # Otherwise, try to build as Example
    return build(node, context)



from dataclasses import dataclass, field, replace

from ruamel import yaml

from ..context import Context
from ..fields import fixed_field
from ..sources import FieldSource, KeySource, ValueSource, YAMLInvalidValue, YAMLValue
from .builders import build_field_source, build_model
from .components import Components
from .components import build as build_components
from .external_documentation import ExternalDocumentation
from .info import Info
from .info import build as build_info
from .path_item import PathItem
from .paths import Paths
from .paths import build as build_paths
from .security_requirement import SecurityRequirement
from .server import Server
from .tag import Tag
from .tag import build as build_tag


__all__ = ["OpenAPI31", "build"]


@dataclass(frozen=True, slots=True)
class OpenAPI31:
    """
    OpenAPI Object representation for OpenAPI 3.1.

    This is the root document object of the OpenAPI document.

    Note: This class uses full duplication (not inheritance from OpenAPI30) to enable precise
    version detection via isinstance() checks. This is critical for validators and converters
    that need to distinguish between OpenAPI 3.0 and 3.1 specifications, as the two versions
    have different validation rules (e.g., paths is required in 3.0 but optional in 3.1).

    Attributes:
        root_node: The top-level node representing the entire OpenAPI object in the original source file
        openapi: REQUIRED. The version number of the OpenAPI Specification that the document uses.
                 Must match the specification version (e.g., "3.1.0", "3.1.1", "3.1.2").
        info: REQUIRED. Provides metadata about the API (title, version, description, etc.)
        json_schema_dialect: The default value for the $schema keyword within Schema Objects in this OAS document.
        servers: An array of Server Objects providing connectivity information to target servers
        paths: The available paths and operations for the API. In OpenAPI 3.1, this is no longer strictly
               required if components or webhooks are present.
        webhooks: The incoming webhooks that MAY be received as part of this API and that the API consumer
                  MAY choose to implement. A map of PathItem Objects.
        components: An element to hold reusable objects for different aspects of the OAS
        security: A declaration of which security mechanisms can be used across the API
        tags: A list of tags used by the specification with additional metadata
        external_docs: Additional external documentation
        extensions: Specification extensions (x-* fields)
    """

    root_node: yaml.Node
    openapi: FieldSource[str] | None = fixed_field()
    info: FieldSource[Info] | None = fixed_field()
    json_schema_dialect: FieldSource[str] | None = fixed_field(
        metadata={"yaml_name": "jsonSchemaDialect"}
    )
    servers: FieldSource[list["Server"]] | None = fixed_field()
    paths: FieldSource[Paths] | None = fixed_field()
    webhooks: FieldSource[dict[KeySource[str], "PathItem"]] | None = fixed_field()
    components: FieldSource[Components] | None = fixed_field()
    security: FieldSource[list["SecurityRequirement"]] | None = fixed_field()
    tags: FieldSource[list[Tag]] | None = fixed_field()
    external_docs: FieldSource["ExternalDocumentation"] | None = fixed_field(
        metadata={"yaml_name": "externalDocs"}
    )
    extensions: dict[KeySource[str], ValueSource[YAMLValue]] = field(default_factory=dict)


def build(
    root: yaml.Node, context: Context | None = None
) -> OpenAPI31 | ValueSource[YAMLInvalidValue]:
    """
    Build an OpenAPI 3.1 object from a YAML node.

    Preserves all source data as-is, regardless of type. This is a low-level/plumbing
    model that provides complete source fidelity for inspection and validation.

    Args:
        root: The YAML node to parse (should be a MappingNode)
        context: Optional parsing context. If None, a default context will be created.

    Returns:
        An OpenAPI31 object if the node is valid, or a ValueSource containing
        the invalid data if the root is not a MappingNode (preserving the invalid data
        and its source location for validation).

    Example:
        from ruamel.yaml import YAML
        yaml = YAML()
        root = yaml.compose('''
        openapi: 3.1.2
        info:
          title: Sample API
          version: 1.0.0
        paths:
          /users:
            get:
              responses:
                '200':
                  description: Success
        webhooks:
          newPet:
            post:
              requestBody:
                description: Pet to add
        ''')
        openapi_doc = build(root)
        assert openapi_doc.openapi.value == '3.1.2'
        assert openapi_doc.info.value.title.value == 'Sample API'
    """
    context = context or Context()

    # Use build_model for initial construction
    openapi_obj = build_model(root, OpenAPI31, context=context)

    # If build_model returned ValueSource (invalid node), return it immediately
    if not isinstance(openapi_obj, OpenAPI31):
        return openapi_obj

    # Manually handle nested complex fields and list fields with custom builders
    replacements = {}
    for key_node, value_node in root.value:
        key = context.yaml_constructor.construct_yaml_str(key_node)

        if key == "info":
            # Handle info field - Info object
            replacements["info"] = FieldSource(
                value=build_info(value_node, context), key_node=key_node, value_node=value_node
            )

        elif key == "paths":
            # Handle paths field - Paths object
            replacements["paths"] = FieldSource(
                value=build_paths(value_node, context), key_node=key_node, value_node=value_node
            )

        elif key == "components":
            # Handle components field - Components object
            replacements["components"] = FieldSource(
                value=build_components(value_node, context),
                key_node=key_node,
                value_node=value_node,
            )

        elif key == "tags":
            # Handle tags field - array of Tag objects
            if isinstance(value_node, yaml.SequenceNode):
                tags_list = []
                for tag_node in value_node.value:
                    tag_obj = build_tag(tag_node, context)
                    tags_list.append(tag_obj)
                replacements["tags"] = FieldSource(
                    value=tags_list, key_node=key_node, value_node=value_node
                )
            else:
                # Not a sequence - preserve as-is for validation
                replacements["tags"] = build_field_source(key_node, value_node, context)

    # Apply all replacements at once
    if replacements:
        openapi_obj = replace(openapi_obj, **replacements)

    return openapi_obj
